package com.jakubu9333.grammarManager.api;

import com.fasterxml.jackson.annotation.JsonView;
import com.jakubu9333.grammarManager.model.FileMetadata;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.View;
import com.jakubu9333.grammarManager.service.FileService;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.CrossOrigin;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.server.ResponseStatusException;

import javax.servlet.http.HttpServletResponse;
import java.util.List;

/**
 * @author Jakub Uhlarik
 */

@RestController
@CrossOrigin(origins = "http://localhost:3000")
@RequestMapping(path = "api/files")
public class FilesControler {
    private final FileService fileService;

    @Autowired
    public FilesControler(FileService fileService) {
        this.fileService = fileService;
    }

    @DeleteMapping("/{id}")
    public ResponseEntity<?> deleteFile(@PathVariable Long id) {
        fileService.deleteFileById(id);
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @DeleteMapping("/grammar/{id}")
    public ResponseEntity<?> deleteFileFromGrammar(@PathVariable Long id) {
        fileService.deleteFilesFromGrammar(id);
        return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    }

    @JsonView(View.All.class)
    @GetMapping("/grammar/{id}")
    public ResponseEntity<List<FileMetadata>> fileHoldersFromGrammar(@PathVariable Long id) {
        return new ResponseEntity<>(fileService.getFilesFromGrammar(id), HttpStatus.OK);
    }

    @GetMapping("/{id}")
    public ResponseEntity<String> getFile(@PathVariable Long id) {
        MediaType typeOfFile;
        Pair<FileMetadata, String> result = fileService.getFileHolderAndStringById(id);
        FileMetadata fileMeta = result.getValue0();
        if (fileMeta.isXml()) {
            typeOfFile = MediaType.APPLICATION_XML;
        } else {
            typeOfFile = MediaType.TEXT_PLAIN;
        }

        return ResponseEntity.ok().contentType(typeOfFile).body(result.getValue1());
    }

    @GetMapping("/{id}/download")
    public ResponseEntity<String> getFileDownload(@PathVariable Long id, HttpServletResponse response) {
        MediaType typeOfFile;
        String fileName;
        Pair<FileMetadata, String> result = fileService.getFileHolderAndStringById(id);
        FileMetadata fileHolder = result.getValue0();
        if (fileHolder.isXml()) {
            fileName = "grammar.xml";
            typeOfFile = MediaType.APPLICATION_XML;
        } else {
            fileName = "grammar.jsgf";
            typeOfFile = MediaType.TEXT_PLAIN;
        }

        response.setHeader("Content-Disposition", "attachment; filename=" + fileName);
        return ResponseEntity.ok().contentType(typeOfFile).body(result.getValue1());
    }


    @GetMapping("{grammarId}/new/{fileFormat}")
    public ResponseEntity<String> fileFromGrammar(@PathVariable String fileFormat, @PathVariable Long grammarId, @RequestParam(defaultValue = "true") boolean save) {
        MediaType typeOfFile;
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> {
                typeOfFile = MediaType.TEXT_PLAIN;
                xml = false;
            }
            case "xml" -> {
                typeOfFile = MediaType.APPLICATION_XML;
                xml = true;
            }
            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Pair<String, Boolean> result = fileService.newFileFromGrammar(grammarId, xml, save);

        return ResponseEntity.status(result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK).contentType(typeOfFile).body(result.getValue0());
    }

    @GetMapping("{grammarId}/new/{fileFormat}/download")
    public ResponseEntity<String> fileFromGrammarDownload(@PathVariable String fileFormat, @PathVariable Long grammarId, HttpServletResponse response, @RequestParam(defaultValue = "true") boolean save) {
        MediaType typeOfFile;
        boolean xml;
        String fileName;
        switch (fileFormat) {
            case "jsgf" -> {
                typeOfFile = MediaType.TEXT_PLAIN;
                xml = false;
                fileName = "grammar.xml";
            }
            case "xml" -> {
                typeOfFile = MediaType.APPLICATION_XML;
                xml = true;
                fileName = "grammar.jsgf";
            }
            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Pair<String, Boolean> result = fileService.newFileFromGrammar(grammarId, xml, save);
        response.setHeader("Content-Disposition", "attachment; filename=" + fileName);
        return ResponseEntity.status(result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK).contentType(typeOfFile).body(result.getValue0());
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("/{fileFormat}")
    public ResponseEntity<Grammar> grammarFromFile(@RequestBody String file, @PathVariable String fileFormat, @RequestParam(defaultValue = "true") boolean save) {
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> xml = false;

            case "xml" -> xml = true;

            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Grammar grammar = fileService.addGrammarFromFile(file, xml, null, false, save).getValue0();
        return new ResponseEntity<>(grammar, HttpStatus.CREATED);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("")
    public ResponseEntity<Grammar> grammarFromFile(@RequestBody String file, @RequestParam(defaultValue = "true") boolean save) {

        Grammar grammar = fileService.addGrammarFromFile(file, null, null, false, save).getValue0();
        return new ResponseEntity<>(grammar, HttpStatus.CREATED);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("u/{id}")
    public ResponseEntity<Grammar> grammarFromFile(@RequestBody String file, @PathVariable Long id, @RequestParam(defaultValue = "false") boolean newGrammar, @RequestParam(defaultValue = "true") boolean save) {
        Pair<Grammar, Boolean> result = fileService.addGrammarFromFile(file, null, id, newGrammar, save);
        if (result == null) {
            return new ResponseEntity<>(HttpStatus.NO_CONTENT);
        }
        return new ResponseEntity<>(result.getValue0(), result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK);
    }

    @JsonView(View.RuleWithoutGrammar.class)
    @PostMapping("{id}/{fileFormat}")
    public ResponseEntity<Grammar> grammarFromFile(@RequestBody String file, @PathVariable String fileFormat, @PathVariable Long id, @RequestParam(defaultValue = "false") boolean newGrammar, @RequestParam(defaultValue = "true") boolean save) {
        boolean xml;
        switch (fileFormat) {
            case "jsgf" -> xml = false;

            case "xml" -> xml = true;

            default -> throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "invalid file format");
        }
        Pair<Grammar, Boolean> result = fileService.addGrammarFromFile(file, xml, id, newGrammar, save);
        if (result == null) {
            return new ResponseEntity<>(HttpStatus.NO_CONTENT);
        }
        return new ResponseEntity<>(result.getValue0(), result.getValue1() ? HttpStatus.CREATED : HttpStatus.OK);
    }
}
