package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.FileMetaDataRepo;
import com.jakubu9333.grammarManager.dao.FileRepository;
import com.jakubu9333.grammarManager.dao.GrammarRepository;
import com.jakubu9333.grammarManager.model.FileContent;
import com.jakubu9333.grammarManager.model.FileMetadata;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.RightSidePart;
import com.jakubu9333.grammarManager.model.Rule;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.transaction.Transactional;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.StringReader;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Jakub Uhlarik
 */
@Service
public class FileService {

    private final FileMetaDataRepo fileMetaDataRepo;

    private final GrammarService grammarService;
    private final GrammarRepository grammarRepository;
    private final FileRepository fileRepository;

    @Autowired
    public FileService(FileMetaDataRepo fileMetaDataRepo, GrammarService grammarService, GrammarRepository grammarRepository, FileRepository fileRepository) {
        this.fileMetaDataRepo = fileMetaDataRepo;
        this.grammarService = grammarService;
        this.grammarRepository = grammarRepository;
        this.fileRepository = fileRepository;
    }


    public static class XmlGrammarHandler extends DefaultHandler {
        private Grammar newGrammar;
        private StringBuilder rightSidePart;

        private Rule rule = null;
        private boolean item = false;

        @Override
        public void startDocument() {
            newGrammar = new Grammar(null, new HashSet<>(), null);
        }

        @Override
        public void characters(char[] ch, int start, int length) {
            if (rightSidePart != null && item) {
                rightSidePart.append(ch, start, length);
            }
        }

        @Override
        public void startElement(String uri, String lName, String qName, Attributes attr) {
            switch (qName) {
                case "grammar" -> {
                    String grammarUri = attr.getValue("xml:base");
                    if (grammarUri != null) {
                        newGrammar.setUri(grammarUri);
                    }
                    String rootLeft = attr.getValue("root");
                    if (rootLeft != null) {
                        newGrammar.getRules().add(new Rule(newGrammar, rootLeft, new HashSet<>(), true));
                    }
                }
                case "rule" -> {
                    rule = new Rule(newGrammar, attr.getValue("id"), new HashSet<>(), false);
                    rightSidePart = new StringBuilder();
                }
                case "item" -> item = true;

                case "ruleref" -> {

                    String spRule = attr.getValue("special");
                    if (spRule != null) {
                        rightSidePart.append("<").append(spRule).append(">");
                    }
                    String ruleRef = attr.getValue("uri");
                    if (ruleRef != null) {
                        if (ruleRef.startsWith("#")) {
                            ruleRef = ruleRef.substring(1);
                        }
                        rightSidePart.append("<").append(ruleRef).append(">");
                    }

                }
            }
        }

        @Override
        public void endElement(String uri, String lName, String qName) {
            switch (qName) {
                case "rule" -> {

                    newGrammar.getRules().add(rule);
                    rule = null;
                }
                case "item" -> {
                    item = false;
                    rule.getRightSide().add(new RightSidePart(rightSidePart.toString()));
                    rightSidePart = new StringBuilder();
                }
            }
        }
    }

    public List<FileMetadata> getFilesFromGrammar(Long id) {
        return fileMetaDataRepo.findAllByGrammarId(id);
    }

    private void deleteSavedContents(FileMetadata fileMetadata) {
        fileRepository.deleteById(fileMetadata.getMongoId());
    }

    @Transactional
    public void deleteFilesFromGrammar(Long grammarId) {
        Grammar grammar = grammarService.getGrammarById(grammarId);
        Set<FileMetadata> files = grammar.getFiles();
        grammar.setFiles(new HashSet<>());
        fileMetaDataRepo.deleteAll(files);
        files.forEach(this::deleteSavedContents);

    }

    @Transactional
    public void deleteFileById(Long id) {
        FileMetadata file = fileMetaDataRepo.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "file:" + id + "not found"));
        file.getGrammar().getFiles().remove(file);
        fileMetaDataRepo.delete(file);
        fileRepository.deleteById(file.getMongoId());
    }

    public List<FileMetadata> getAllFiles() {
        return fileMetaDataRepo.findAll();
    }


    public Pair<FileMetadata, String> getFileHolderAndStringById(Long id) {
        FileMetadata fileMetadata = fileMetaDataRepo.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "file id:" + id + " not found"));
        return new Pair<>(fileMetadata, getStringFromFileHolder(fileMetadata));
    }


    private Grammar parseFile(String file, boolean xml) {
        Grammar grammar;
        if (xml) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();

                XmlGrammarHandler handler = new XmlGrammarHandler();
                InputSource is = new InputSource(new StringReader(file));
                saxParser.parse(is, handler);

                grammar = handler.newGrammar;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "parsing error", e);
            }

        } else {
            grammar = jsgfParse(file);
        }
        return grammarService.preProcessGrammar(grammar);
    }

    private String saveFile(String content) {
        FileContent file = new FileContent(content);
        return fileRepository.insert(file).getId();
    }

    private boolean getFileExtension(String file) {
        Pattern xml = Pattern.compile("^<\\?xml.*\\?>");
        Matcher mXml = xml.matcher(file);
        if (mXml.find()) {
            return true;
        }
        Pattern jsgf = Pattern.compile("^#JSGF.*;");
        Matcher mJsgf = jsgf.matcher(file);
        if (mJsgf.find()) {
            return false;
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "format cant be found from file");
    }

    @Transactional
    public Pair<Grammar, Boolean> addGrammarFromFile(String file, Boolean xml, Long id, boolean generateNewOnConflict, boolean saveFile) {
        if (xml == null) {
            xml = getFileExtension(file);
        }
        Grammar grammar = parseFile(file, xml);

        FileMetadata fileHolder = new FileMetadata(xml, false, null);
        if (saveFile) {
            fileMetaDataRepo.save(fileHolder);
            fileHolder.setSize(file.length());
        }

        if (id == null || !grammarRepository.existsById(id)) {
            if (saveFile) {
                fileHolder.setMongoId(saveFile(file));
                grammar.getFiles().add(fileHolder);
                fileHolder.setGrammar(grammar);
            }
            return new Pair<>(grammarService.addGrammar(grammar), true);

        }
        Grammar oldGrammar = grammarService.getGrammarById(id);
        if (oldGrammar.equals(grammar)) {
            if (saveFile) {
                fileHolder.setMongoId(saveFile(file));
                oldGrammar.getFiles().add(fileHolder);
                fileHolder.setGrammar(oldGrammar);
            }
            return new Pair<>(oldGrammar, false);
        }
        if (generateNewOnConflict) {
            if (saveFile) {
                fileHolder.setMongoId(saveFile(file));
                grammar.getFiles().add(fileHolder);
                fileHolder.setGrammar(grammar);
            }
            return new Pair<>(grammarService.addGrammar(grammar), true);
        }
        if (saveFile) {
            fileMetaDataRepo.delete(fileHolder);
        }

        return null;
    }


    private Grammar jsgfParse(String file) {
        //todo jsgf parse + kleene star + angle brackets
        Grammar grammar = new Grammar();
        boolean comment = false;
        boolean complexCommentNow;
        boolean start = false;
        String[] lines = file.split("[;\n]");
        String uncommentedLine = null;
        int i = 1;
        boolean root = false;
        boolean setRoot = false;
        for (String line : lines) {
            complexCommentNow = false;
            if (setRoot) {
                root = true;
                setRoot = false;
            }
            line = line.trim();

            if (line.contains("/*")) {
                String[] splitByComment = line.split("/\\*");
                uncommentedLine = splitByComment[0];
                comment = true;
                complexCommentNow = true;
            }
            if (line.contains("*/")) {
                if (comment) {
                    comment = false;
                } else {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "comment without start; line= " + i);
                }
                String[] splitByComment = line.split("\\*/");
                if (splitByComment.length > 0) {
                    line = splitByComment[1];
                }

                if (complexCommentNow) {
                    line = uncommentedLine + line;
                }


            }
            if (line.contains("//")) {
                String[] splitByComment = line.split("//");
                line = splitByComment[1];
                if (line.equals("root")) {
                    setRoot = true;
                }
                continue;
            }
            if (line.matches("#JSGF V[0-9].[0-9]+")) {
                start = true;
                i++;
                continue;
            }
            if (line.contains("grammar ") & !comment) {
                if (grammar.getUri() != null) {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "duplicate grammar name declaration; line=" + i);
                }
                String[] splitLine = line.split(" ");
                grammar.setUri(splitLine[1]);
            }
            if (line.contains("=") & !comment) {
                //rule
                String[] splitLine = line.split("=");
                String leftSymbol = splitLine[0].trim();
                RightSidePart rightSidePart = new RightSidePart(splitLine[1].trim());
                Rule rule = new Rule(null, grammar, leftSymbol, Set.of(rightSidePart), root);
                root = false;
                grammar.getRules().add(rule);
            }
            i++;
        }
        return grammar;
    }

    private void printRightSide(StringBuilder sb, String rightSide) {
        if (rightSide.contains("<") && rightSide.contains(">")) {

            int leftindex = 0;
            int rightindex = 0;

            int begin = 0;
            while (leftindex >= 0 && rightindex >= 0) {
                leftindex = rightSide.indexOf('<', rightindex);
                rightindex = rightSide.indexOf('>', leftindex);
                //between two rules or begin
                if (leftindex - begin > 0) {

                    sb.append(rightSide, begin, leftindex);
                }
                //after rules
                if (leftindex == -1 && rightSide.length() > begin) {
                    sb.append(rightSide.substring(begin));
                }
                if (leftindex >= 0 && rightindex >= 0) {
                    String ruleInside = rightSide.substring(leftindex + 1, rightindex);
                    if (ruleInside.equals("NULL")) {
                        sb.append("<ruleref special=\"NULL\"/>");
                    } else if (ruleInside.equals("VOID")) {
                        sb.append("<ruleref special=\"VOID\"/>");
                    } else {
                        sb.append("<ruleref uri=\"#").append(ruleInside).append("\"/>");
                    }

                }
                begin = rightindex + 1;

            }
            return;
        }
        if (!rightSide.contains("<") && !rightSide.contains(">")) {
            sb.append(rightSide).append("\n");
            return;
        }
        throw new ResponseStatusException(HttpStatus.EXPECTATION_FAILED, "bad rule format");
    }

    private void writeRule(StringBuilder sb, Rule rule, boolean xml) {
        if (!xml) {
            if (rule.getRoot()) {
                sb.append("//root;\n");
            }
            int i = 0;
            StringBuilder rightSide = new StringBuilder();
            Set<RightSidePart> setRightSide = rule.getRightSide();
            for (RightSidePart part : setRightSide) {
                rightSide.append(part.getPart());
                if (i + 1 < setRightSide.size()) {
                    rightSide.append(" | ");
                }
                i++;
            }
            sb.append(rule.getLeftSymbol()).append(" = ").append(rightSide).append(";\n");
            return;
        }
        sb.append("<rule id=\"");
        sb.append(nameFromLeftSymbol(rule)).append("\">\n");
        Set<RightSidePart> rightSide = rule.getRightSide();
        if (rightSide.size() > 1) {
            sb.append("<one-of>\n");
            for (RightSidePart part : rightSide) {
                sb.append("<item>\n");
                printRightSide(sb, part.getPart());
                sb.append("</item>\n");
            }
            sb.append("</one-of>\n");
        } else {
            sb.append("<item>\n");
            printRightSide(sb, rightSide.stream().map(RightSidePart::getPart).toList().get(0));
            sb.append("</item>\n");
        }
        sb.append("</rule>\n");
    }

    private String nameFromLeftSymbol(Rule rule) {
        String oldLeftSymbol = rule.getLeftSymbol();
        return rule.getLeftSymbol().substring(1, oldLeftSymbol.length() - 1);
    }

    private void writeHeader(StringBuilder sb, Grammar grammar, boolean xml) {
        //jsgf
        if (!xml) {
            sb.append("#JSGF v 1.0;\n");
            sb.append("grammar ").append(grammar.getUri()).append(";\n\n");
            return;
        }
        //xml
        sb.append("<?xml version=\"1.0\"?>\n");
        sb.append("<grammar version=\"1.0\"\n");
        Rule rootRule;
        String uri;
        if ((uri = grammar.getUri()) != null) {
            sb.append("xml:base=\"").append(uri).append("\"\n");
        }
        if ((rootRule = grammar.getRootRule()) != null) {
            sb.append("root=\"").append(nameFromLeftSymbol(rootRule)).append("\"\n");
        }

        sb.append(" xmlns=\"http://www.w3.org/2001/06/grammar\">\n");
    }

    private String generateFile(Grammar grammar, boolean xml) {
        StringBuilder sb = new StringBuilder();
        writeHeader(sb, grammar, xml);
        Rule rootRule;
        if (null != (rootRule = grammar.getRootRule())) {
            writeRule(sb, rootRule, xml);
        }
        for (Rule rule : grammar.getRules()) {
            if (!rule.getRoot()) {
                writeRule(sb, rule, xml);
            }
        }
        if (xml) {
            sb.append("</grammar>");
        }
        return sb.toString();
    }

    public String getStringFromFileHolder(FileMetadata fileData) {

        FileContent fc = fileRepository.findById(fileData.getMongoId()).orElse(null);
        return fc != null ? fc.getContent() : null;
    }

    @Transactional
    public Pair<String, Boolean> newFileFromGrammar(Long id, boolean xml, boolean save) {
        Grammar grammar = grammarService.getGrammarById(id);
        FileMetadata fileHolder = fileMetaDataRepo.getFileHolderByGeneratedTrueAndGrammar_IdAndXml(id, xml);

        if (fileHolder != null && (grammar.getUpdateTime() == null || fileHolder.getCreateTime().compareTo(grammar.getUpdateTime()) >= 0)) {
            String content = getStringFromFileHolder(fileHolder);
            return new Pair<>(content, false);
        }
        if (fileHolder != null) {
            grammar.getFiles().remove(fileHolder);
            fileMetaDataRepo.delete(fileHolder);
            fileRepository.deleteById(fileHolder.getMongoId());
        }

        String fileContent = generateFile(grammar, xml);
        if (!save) {
            return new Pair<>(fileContent, false);
        }
        fileHolder = new FileMetadata(xml, true, grammar);
        fileMetaDataRepo.save(fileHolder);
        grammar.getFiles().add(fileHolder);

        fileHolder.setSize(fileContent.length());
        fileHolder.setMongoId(saveFile(fileContent));
        return new Pair<>(fileContent, true);
    }


}

